# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.server` module`
"""

from unittest import TestCase
import os
from os import path
import io
import stat
import time
from random import SystemRandom
import socket
import ssl
import json
from hashlib import sha1
import struct

from .helpers import TempDir
import degu
from degu.sslhelpers import random_id
from degu.misc import TempPKI, TempServer, TempSSLServer
from degu.client import Client, SSLClient, build_client_sslctx
from degu.base import _TYPE_ERROR
from degu import rgi, base, server


random = SystemRandom()


def standard_harness_app(session, request, bodies):
    if len(request.path) == 3 and request.path[0] == 'status':
        code = int(request.path[1])
        reason = request.path[2]
        return (code, reason, {}, None)


class TestFunctions(TestCase):
    def test_build_server_sslctx(self):
        # Bad sslconfig type:
        with self.assertRaises(TypeError) as cm:
            server.build_server_sslctx('bad')
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('sslconfig', dict, str, 'bad')
        )

        # The remaining test both build_server_sslctx() directly, and the
        # pass-through from _validate_server_sslctx():
        server_sslctx_funcs = (
            server.build_server_sslctx,
            server._validate_server_sslctx,
        )

        # Non absulute, non normalized paths:
        good = {
            'cert_file': '/my/server.cert',
            'key_file': '/my/server.key',
            'ca_file': '/my/client.ca',
            'ca_path': '/my/client.ca.dir',
        }
        for key in good.keys():
            for func in server_sslctx_funcs:
                # Relative path:
                bad = good.copy()
                value = 'relative/path'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

                # Non-normalized path with directory traversal:
                bad = good.copy()
                value = '/my/../secret/path'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

                # Non-normalized path with trailing slash:
                bad = good.copy()
                value = '/sorry/very/strict/'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )


        pki = TempPKI(client_pki=True)

        # Typical sslconfig with client authentication:
        self.assertEqual(set(pki.server_sslconfig),
            {'cert_file', 'key_file', 'ca_file'}
        )
        for func in server_sslctx_funcs:
            sslctx = func(pki.server_sslconfig)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertTrue(sslctx.options & ssl.OP_NO_COMPRESSION)
            self.assertTrue(sslctx.options & ssl.OP_SINGLE_ECDH_USE)
            self.assertTrue(sslctx.options & ssl.OP_CIPHER_SERVER_PREFERENCE)

        # New in Degu 0.3: should not be able to accept connections from
        # unauthenticated clients by merely omitting ca_file/ca_path:
        sslconfig = pki.server_sslconfig
        del sslconfig['ca_file']
        for func in server_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func(sslconfig)
            self.assertEqual(str(cm.exception),
                'must include ca_file or ca_path (or allow_unauthenticated_clients)'
            )

        # Typical config allowing anonymous clients:
        sslconfig['allow_unauthenticated_clients'] = True
        for func in server_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_NONE)
            self.assertTrue(sslctx.options & ssl.OP_NO_COMPRESSION)
            self.assertTrue(sslctx.options & ssl.OP_SINGLE_ECDH_USE)
            self.assertTrue(sslctx.options & ssl.OP_CIPHER_SERVER_PREFERENCE)

        # Cannot mix ca_file/ca_path with allow_unauthenticated_clients:
        sslconfig['ca_file'] = '/my/client.ca'
        for func in server_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func(sslconfig)
            self.assertEqual(str(cm.exception),
                'ca_file/ca_path with allow_unauthenticated_clients'
            )
        sslconfig['ca_path'] = sslconfig.pop('ca_file')
        for func in server_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func(sslconfig)
            self.assertEqual(str(cm.exception),
                'ca_file/ca_path with allow_unauthenticated_clients'
            )

        # True is only allowed value for allow_unauthenticated_clients:
        sslconfig.pop('ca_path')
        for bad in (1, 0, False, None):
            sslconfig['allow_unauthenticated_clients'] = bad
            for func in server_sslctx_funcs:
                with self.assertRaises(ValueError) as cm:
                    func(sslconfig)
                self.assertEqual(str(cm.exception),
                    'True is only allowed value for allow_unauthenticated_clients'
                )

    def test__validate_server_sslctx(self):
        # Bad type:
        with self.assertRaises(TypeError) as cm:
            server._validate_server_sslctx(ssl.SSLContext)
        self.assertEqual(str(cm.exception), 'sslctx must be an ssl.SSLContext')

        # Wrong protocol:
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(ssl.SSLContext(ssl.PROTOCOL_TLSv1))
        self.assertEqual(str(cm.exception),
            'sslctx.protocol must be ssl.PROTOCOL_TLSv1_2'
        )

        # Don't allow ssl.CERT_OPTIONAL:
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        sslctx.verify_mode = ssl.CERT_OPTIONAL
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.verify_mode cannot be ssl.CERT_OPTIONAL'
        )

        # options missing OP_NO_COMPRESSION:
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self.assertIs(sslctx.verify_mode, ssl.CERT_NONE)
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_NO_COMPRESSION'
        )

        # options missing OP_SINGLE_ECDH_USE:
        sslctx.options |= ssl.OP_NO_COMPRESSION
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_SINGLE_ECDH_USE'
        )

        # options missing OP_CIPHER_SERVER_PREFERENCE:
        sslctx.options |= ssl.OP_SINGLE_ECDH_USE
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_CIPHER_SERVER_PREFERENCE'
        )

        # All good:
        sslctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE
        self.assertIs(server._validate_server_sslctx(sslctx), sslctx)

        # Now again, this time with CERT_REQUIRED:
        # options missing OP_NO_COMPRESSION:
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        sslctx.verify_mode = ssl.CERT_REQUIRED
        self.assertIs(sslctx.verify_mode, ssl.CERT_REQUIRED)
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_NO_COMPRESSION'
        )

        # options missing OP_SINGLE_ECDH_USE:
        sslctx.options |= ssl.OP_NO_COMPRESSION
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_SINGLE_ECDH_USE'
        )

        # options missing OP_CIPHER_SERVER_PREFERENCE:
        sslctx.options |= ssl.OP_SINGLE_ECDH_USE
        with self.assertRaises(ValueError) as cm:
            server._validate_server_sslctx(sslctx)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_CIPHER_SERVER_PREFERENCE'
        )

        # All good:
        sslctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE
        self.assertIs(server._validate_server_sslctx(sslctx), sslctx)

    def test_get_credentials(self):
        class MockSocket:
            def __init__(self, cred):
                self._cred = cred
                self._calls = []

            def setsockopt(self, *args):
                self._calls.append(('setsockopt',) + args)

            def getsockopt(self, *args):
                self._calls.append(('getsockopt',) + args)
                return struct.pack('3i', *self._cred)

        cred = (9999, 8888, 7777)
        sock = MockSocket(cred)
        self.assertEqual(server._get_credentials(sock), cred)
        size = struct.calcsize('3i')
        self.assertEqual(sock._calls, [
            ('setsockopt', socket.SOL_SOCKET, socket.SO_PASSCRED, 1),
            ('getsockopt', socket.SOL_SOCKET, socket.SO_PEERCRED, size)
        ])

    def test_shutdown_and_close(self):
        class MockSocket:
            def __init__(self, exc=None):
                self._exc = exc
                self._calls = []

            def shutdown(self, how):
                self._calls.append(('shutdown', how))
                if self._exc is not None:
                    raise self._exc

            def close(self):
                self._calls.append('close')

        sock = MockSocket()
        self.assertIsNone(server._shutdown_and_close(sock))
        self.assertEqual(sock._calls, [
            ('shutdown', socket.SHUT_RDWR),
            'close',
        ])

        sock = MockSocket(ConnectionError(random_id()))
        self.assertIsNone(server._shutdown_and_close(sock))
        self.assertEqual(sock._calls, [
            ('shutdown', socket.SHUT_RDWR),
            'close',
        ])

        marker = random_id()
        exc = ValueError(marker)
        sock = MockSocket(exc)
        with self.assertRaises(ValueError) as cm:
            server._shutdown_and_close(sock)
        self.assertIs(cm.exception, exc)
        self.assertEqual(str(cm.exception), marker)
        self.assertEqual(sock._calls, [
            ('shutdown', socket.SHUT_RDWR),
        ]) 


class BadApp:
    """
    Not callable.
    """


def good_app(session, request, bodies):
    return (200, 'OK', {}, None)


class BadConnectionHandler:
    def __call__(self, session, request, bodies):
        pass

    on_connect = 'nope'


class GoodConnectionHandler:
    def __call__(self, session, request, bodies):
        pass

    def on_connect(self, session, sock):
        pass


class TestServer(TestCase):
    def test_init(self):
        # Bad address type:
        with self.assertRaises(TypeError) as cm:
            server.Server(1234, good_app)
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('address', (tuple, str, bytes), int, 1234)
        )

        # Wrong number of items in address tuple:
        bad_addresses = [
            ('::1',),
            ('127.0.0.1',),
            ('::1', 0, 0),
            ('127.0.0.1', 0, 0),
            ('::1', 0, 0, 0, 0),
            ('127.0.0.1', 0, 0, 0, 0),
        ]
        for address in bad_addresses:
            self.assertIn(len(address), {1, 3, 5})
            with self.assertRaises(ValueError) as cm:
                server.Server(address, good_app)
            self.assertEqual(str(cm.exception),
                'address: must have 2 or 4 items; got {!r}'.format(address)
            )

        # Non-normalized socket filename:
        with self.assertRaises(ValueError) as cm:
            server.Server('foo', good_app)
        self.assertEqual(str(cm.exception),
            "address: bad socket filename: 'foo'"
        )

        # app not callable:
        bad_app = BadApp()
        with self.assertRaises(TypeError) as cm:
            server.Server(degu.IPv6_LOOPBACK, bad_app)
        self.assertEqual(str(cm.exception),
            'app: not callable: {!r}'.format(bad_app)
        )

        # app.on_connect not callable:
        bad_app = BadConnectionHandler()
        with self.assertRaises(TypeError) as cm:
            server.Server(degu.IPv6_LOOPBACK, bad_app)
        self.assertEqual(str(cm.exception),
            'app.on_connect: not callable: {!r}'.format(bad_app)
        )

        # Good app.on_connect:
        app = GoodConnectionHandler()
        inst = server.Server(degu.IPv6_LOOPBACK, app)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('::1', port, 0, 0))
        self.assertIs(inst.app, app)

        # IPv6 loopback:
        inst = server.Server(degu.IPv6_LOOPBACK, good_app)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('::1', port, 0, 0))
        self.assertIs(inst.app, good_app)

        # IPv6 any:
        inst = server.Server(degu.IPv6_ANY, good_app)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('::', port, 0, 0))
        self.assertIs(inst.app, good_app)

        # IPv4 loopback:
        inst = server.Server(degu.IPv4_LOOPBACK, good_app)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('127.0.0.1', port))
        self.assertIs(inst.app, good_app)

        # IPv4 any:
        inst = server.Server(degu.IPv4_ANY, good_app)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('0.0.0.0', port))
        self.assertIs(inst.app, good_app)

        # Socket filename:
        tmp = TempDir()
        filename = tmp.join('my.socket')
        self.assertFalse(path.exists(filename))
        inst = server.Server(filename, good_app)
        self.assertIsInstance(inst.sock, socket.socket)
        self.assertEqual(inst.address, filename)
        self.assertEqual(inst.sock.getsockname(), filename)
        self.assertIs(inst.app, good_app)
        self.assertTrue(stat.S_ISSOCK(os.stat(filename).st_mode))

        # Linux abstract socket names:
        inst = server.Server(b'', good_app)
        self.assertIsInstance(inst.sock, socket.socket)
        self.assertEqual(inst.address, inst.sock.getsockname())
        self.assertIsInstance(inst.address, bytes)
        self.assertIs(inst.app, good_app)

        # Test options:
        inst = server.Server(degu.IPv6_LOOPBACK, good_app)
        self.assertEqual(inst.timeout, 30)
        self.assertEqual(inst.max_connections, 50)
        self.assertEqual(inst.max_requests, 1000)
        self.assertEqual(inst.options, {})
        options = {
            'timeout': 1,
            'max_connections': 2,
            'max_requests': 3,  
        }
        for (key, val) in options.items():
            kw = {key: val}
            inst = server.Server(degu.IPv6_LOOPBACK, good_app, **kw)
            self.assertEqual(getattr(inst, key), val)
            self.assertEqual(inst.options, kw)

        # Test unsupported options:
        with self.assertRaises(TypeError) as cm:
            server.Server(degu.IPv6_LOOPBACK, good_app, foo=17)
        self.assertEqual(str(cm.exception),
            'unsupported Server() **options: foo'
        )
        with self.assertRaises(TypeError) as cm:
            server.Server(degu.IPv6_LOOPBACK, good_app, foo=17, bar=19)
        self.assertEqual(str(cm.exception),
            'unsupported Server() **options: bar, foo'
        )

    def test_repr(self):
        pairs = (
            (good_app, 'degu.tests.test_server.good_app'),
            (rgi.Validator, 'degu.rgi.Validator'),
            (rgi.Validator(good_app), 'degu.rgi.Validator(<...>)'),
        )

        class Custom(server.Server):
            pass    

        for (app, fqname) in pairs:
            inst = server.Server(degu.IPv6_LOOPBACK, app)
            self.assertEqual(repr(inst),
                'Server({!r}, {})'.format(inst.address, fqname)
            )
            inst = Custom(degu.IPv6_LOOPBACK, app)
            self.assertEqual(repr(inst),
                'Custom({!r}, {})'.format(inst.address, fqname)
            )


class TestSSLServer(TestCase):
    def test_init(self):
        # sslctx is not an ssl.SSLContext:
        with self.assertRaises(TypeError) as cm:
            server.SSLServer('foo', degu.IPv6_LOOPBACK, good_app)
        self.assertEqual(str(cm.exception), 'sslctx must be an ssl.SSLContext')

        # Bad SSL protocol version:
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        with self.assertRaises(ValueError) as cm:
            server.SSLServer(sslctx, degu.IPv6_LOOPBACK, good_app)
        self.assertEqual(str(cm.exception),
            'sslctx.protocol must be ssl.PROTOCOL_TLSv1_2'
        )

        # Note: Python 3.3.4 (and presumably 3.4.0) now disables SSLv2 by
        # default (which is good); Degu enforces this (also good), but because
        # we cannot unset the ssl.OP_NO_SSLv2 bit, we can't unit test to check
        # that Degu enforces this, so for now, we set the bit here so it works
        # with Python 3.3.3 still; see: http://bugs.python.org/issue20207
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        sslctx.options |= ssl.OP_NO_SSLv2

        # not (options & ssl.OP_NO_COMPRESSION)
        sslctx.options |= ssl.OP_NO_SSLv2
        with self.assertRaises(ValueError) as cm:
            server.SSLServer(sslctx, '::1', good_app)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_NO_COMPRESSION'
        )

        # not (options & ssl.OP_SINGLE_ECDH_USE)
        sslctx.options |= ssl.OP_NO_COMPRESSION
        with self.assertRaises(ValueError) as cm:
            server.SSLServer(sslctx, '::1', good_app)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_SINGLE_ECDH_USE'
        )

        # not (options & ssl.OP_CIPHER_SERVER_PREFERENCE)
        sslctx.options |= ssl.OP_SINGLE_ECDH_USE
        with self.assertRaises(ValueError) as cm:
            server.SSLServer(sslctx, '::1', good_app)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_CIPHER_SERVER_PREFERENCE'
        )

        # Good sslctx from here on:
        sslctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE

        # Bad address type:
        with self.assertRaises(TypeError) as cm:
            server.SSLServer(sslctx, 1234, good_app)
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('address', (tuple, str, bytes), int, 1234)
        )

        # Wrong number of items in address tuple:
        bad_addresses = [
            ('::1',),
            ('127.0.0.1',),
            ('::1', 0, 0),
            ('127.0.0.1', 0, 0),
            ('::1', 0, 0, 0, 0),
            ('127.0.0.1', 0, 0, 0, 0),
        ]
        for address in bad_addresses:
            self.assertIn(len(address), {1, 3, 5})
            with self.assertRaises(ValueError) as cm:
                server.SSLServer(sslctx, address, good_app)
            self.assertEqual(str(cm.exception),
                'address: must have 2 or 4 items; got {!r}'.format(address)
            )

        # app not callable:
        bad_app = BadApp()
        with self.assertRaises(TypeError) as cm:
            server.SSLServer(sslctx, degu.IPv6_LOOPBACK, bad_app)
        self.assertEqual(str(cm.exception),
            'app: not callable: {!r}'.format(bad_app)
        )

        # app.on_connect not callable:
        bad_app = BadConnectionHandler()
        with self.assertRaises(TypeError) as cm:
            server.SSLServer(sslctx, degu.IPv6_LOOPBACK, bad_app)
        self.assertEqual(str(cm.exception),
            'app.on_connect: not callable: {!r}'.format(bad_app)
        )

        # Good app.on_connect:
        app = GoodConnectionHandler()
        inst = server.SSLServer(sslctx, degu.IPv6_LOOPBACK, app)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('::1', port, 0, 0))
        self.assertIs(inst.app, app)

        # IPv6 loopback:
        inst = server.SSLServer(sslctx, degu.IPv6_LOOPBACK, good_app)
        self.assertIs(inst.sslctx, sslctx)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('::1', port, 0, 0))
        self.assertIs(inst.app, good_app)

        # IPv6 any:
        inst = server.SSLServer(sslctx, degu.IPv6_ANY, good_app)
        self.assertIs(inst.sslctx, sslctx)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('::', port, 0, 0))
        self.assertIs(inst.app, good_app)

        # IPv4 loopback:
        inst = server.SSLServer(sslctx, degu.IPv4_LOOPBACK, good_app)
        self.assertIs(inst.sslctx, sslctx)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('127.0.0.1', port))
        self.assertIs(inst.app, good_app)

        # IPv4 any:
        inst = server.SSLServer(sslctx, degu.IPv4_ANY, good_app)
        self.assertIs(inst.sslctx, sslctx)
        self.assertIsInstance(inst.sock, socket.socket)
        port = inst.sock.getsockname()[1]
        self.assertEqual(inst.address, ('0.0.0.0', port))
        self.assertIs(inst.app, good_app)

        # Test options:
        inst = server.SSLServer(sslctx, degu.IPv6_LOOPBACK, good_app)
        self.assertEqual(inst.timeout, 30)
        self.assertEqual(inst.max_connections, 50)
        self.assertEqual(inst.max_requests, 1000)
        self.assertEqual(inst.options, {})
        options = {
            'timeout': 1,
            'max_connections': 2,
            'max_requests': 3,  
        }
        for (key, val) in options.items():
            kw = {key: val}
            inst = server.SSLServer(sslctx, degu.IPv6_LOOPBACK, good_app, **kw)
            self.assertEqual(getattr(inst, key), val)
            self.assertEqual(inst.options, kw)

        # Test unsupported options:
        with self.assertRaises(TypeError) as cm:
            server.SSLServer(sslctx, degu.IPv6_LOOPBACK, good_app, foo=17)
        self.assertEqual(str(cm.exception),
            'unsupported SSLServer() **options: foo'
        )
        with self.assertRaises(TypeError) as cm:
            server.SSLServer(sslctx, degu.IPv6_LOOPBACK, good_app, foo=17, bar=19)
        self.assertEqual(str(cm.exception),
            'unsupported SSLServer() **options: bar, foo'
        )

    def test_repr(self):
        pairs = (
            (good_app, 'degu.tests.test_server.good_app'),
            (rgi.Validator, 'degu.rgi.Validator'),
            (rgi.Validator(good_app), 'degu.rgi.Validator(<...>)'),
        )
        pki = TempPKI()
        sslctx = server.build_server_sslctx(pki.server_sslconfig)

        class Custom(server.SSLServer):
            pass    

        for (app, fqname) in pairs:
            inst = server.SSLServer(sslctx, degu.IPv6_LOOPBACK, app)
            self.assertEqual(repr(inst),
                'SSLServer(<sslctx>, {!r}, {})'.format(inst.address, fqname)
            )
            inst = Custom(sslctx, degu.IPv6_LOOPBACK, app)
            self.assertEqual(repr(inst),
                'Custom(<sslctx>, {!r}, {})'.format(inst.address, fqname)
            )


CHUNKS = []
for i in range(7):
    size = random.randint(1, 5000)
    CHUNKS.append(os.urandom(size))
CHUNKS.append(b'')
CHUNKS = tuple(CHUNKS)
wfile = io.BytesIO()
for data in CHUNKS:
    base.write_chunk(wfile, (None, data))
ENCODED_CHUNKS = wfile.getvalue()
del wfile


def chunked_request_app(session, request, bodies):
    assert request.method == 'POST'
    assert request.mount == []
    assert request.path == []
    assert isinstance(request.body, bodies.ChunkedBody)
    assert request.headers['transfer-encoding'] == 'chunked'
    result = []
    for (extension, data) in request.body:
        result.append(sha1(data).hexdigest())
    body = json.dumps(result).encode('utf-8')
    headers = {'content-length': len(body), 'content-type': 'application/json'}
    return (200, 'OK', headers, body)


def chunked_response_app(session, request, bodies):
    assert request.method == 'GET'
    assert request.mount == []
    assert request.body is None
    headers = {'transfer-encoding': 'chunked'}
    if request.path == ['foo']:
        rfile = io.BytesIO(ENCODED_CHUNKS)
    elif request.path == ['bar']:
        rfile = io.BytesIO(b'0\r\n\r\n')
    else:
        return (404, 'Not Found', {}, None)
    body = bodies.ChunkedBody(rfile)
    return (200, 'OK', headers, body)


DATA1 = os.urandom(1776)
DATA2 = os.urandom(3469)
DATA = DATA1 + DATA2


def response_app(session, request, bodies):
    assert request.method == 'GET'
    assert request.mount == []
    assert request.body is None
    if request.path == ['foo']:
        body = bodies.Body(io.BytesIO(DATA), len(DATA))
    elif request.path == ['bar']:
        body = bodies.Body(io.BytesIO(), 0)
    else:
        return (404, 'Not Found', {}, None)
    headers = {'content-length': body.content_length}
    return (200, 'OK', headers, body)


def timeout_app(session, request, bodies):
    assert request.method == 'POST'
    assert request.mount == []
    assert request.body is None
    if request.path == ['foo']:
        # Used to test timeout on server side:
        return (200, 'OK', {}, None)
    if request.path == ['bar']:
        # Used to test timeout on client side:
        #time.sleep(CLIENT_SOCKET_TIMEOUT + 2)
        return (400, 'Bad Request', {}, None)
    return (404, 'Not Found', {}, None)


class AppWithConnectionHandler:
    def __init__(self,  marker, accept):
        assert isinstance(marker, bytes) and len(marker) > 0
        assert isinstance(accept, bool)
        self.marker = marker
        self.accept = accept

    def __call__(self, session, request, bodies):
        return (200, 'OK', {}, self.marker)

    def on_connect(self, session, sock):
        return self.accept



class TestLiveServer(TestCase):
    address = degu.IPv4_LOOPBACK

    def build_with_app(self, app, **options):
        httpd = TempServer(self.address, rgi.Validator(app), **options)
        client = Client(httpd.address)
        return (httpd, client)

    def test_timeout(self):
        """
        Do a realistic live socket timeout test.

        This is a very important test that we absolutely want to run during
        package builds on the build servers, despite the fact that it's a
        rather time consuming test.

        However, during day to day development, it's often handy to skip the
        timeout tests for the sake of getting quicker feedback on a code change.

        You can run all the unit tests minus the timeout tests like this::

            $ ./setup.py test --skip-slow

        You can also accomplish the same with an environment variable::

            $ DEGU_TEST_SKIP_SLOW=true ./setup.py test

        """
        if os.environ.get('DEGU_TEST_SKIP_SLOW') == 'true':
            self.skipTest('skipping as DEGU_TEST_SKIP_SLOW is set')

        timeout = 3
        (httpd, client) = self.build_with_app(timeout_app, timeout=timeout)
        conn = client.connect()

        # Make an inital request:
        self.assertEqual(conn.request('POST', '/foo', {}, None), 
            (200, 'OK', {}, None)
        )

        # Wait till 1 second *before* the timeout should happen, to make sure
        # connection is still open:
        time.sleep(timeout - 0.5)
        self.assertEqual(conn.request('POST', '/foo', {}, None),
            (200, 'OK', {}, None)
        )

        # Now wait till 1 second *after* the timeout should have happened, to
        # make sure the connection was closed by the server:
        time.sleep(timeout + 0.5)
        with self.assertRaises(ConnectionError):
            conn.request('POST', '/foo', {}, None)
        self.assertIs(conn.closed, True)
        conn = client.connect()
        self.assertEqual(conn.request('POST', '/foo', {}, None),
            (200, 'OK', {}, None)
        )
        httpd.terminate()

    def test_chunked_request(self):
        (httpd, client) = self.build_with_app(chunked_request_app)
        conn = client.connect()

        body = base.bodies.ChunkedBody(io.BytesIO(ENCODED_CHUNKS))
        response = conn.request('POST', '/', {}, body)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers,
            {'content-length': 352, 'content-type': 'application/json'}
        )
        self.assertIsInstance(response.body, base.bodies.Body)
        self.assertEqual(json.loads(response.body.read().decode('utf-8')),
            [sha1(chunk).hexdigest() for chunk in CHUNKS]
        )

        body = base.bodies.ChunkedBody(io.BytesIO(b'0\r\n\r\n'))
        response = conn.request('POST', '/', {}, body)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers,
            {'content-length': 44, 'content-type': 'application/json'}
        )
        self.assertIsInstance(response.body, base.bodies.Body)
        self.assertEqual(json.loads(response.body.read().decode('utf-8')),
            [sha1(b'').hexdigest()]
        )

        httpd.terminate()

    def test_chunked_response(self):
        (httpd, client) = self.build_with_app(chunked_response_app)
        conn = client.connect()

        response = conn.request('GET', '/foo', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'transfer-encoding': 'chunked'})
        self.assertIsInstance(response.body, base.bodies.ChunkedBody)
        self.assertEqual(tuple(response.body),
            tuple((None, data) for data in CHUNKS)
        )

        response = conn.request('GET', '/bar', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'transfer-encoding': 'chunked'})
        self.assertIsInstance(response.body, base.bodies.ChunkedBody)
        self.assertEqual(list(response.body), [(None, b'')])

        response = conn.request('GET', '/baz', {}, None)
        self.assertEqual(response.status, 404)
        self.assertEqual(response.reason, 'Not Found')
        self.assertEqual(response.headers, {})
        self.assertIsNone(response.body)

        response = conn.request('GET', '/foo', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'transfer-encoding': 'chunked'})
        self.assertIsInstance(response.body, base.bodies.ChunkedBody)
        self.assertEqual(tuple(response.body),
            tuple((None, data) for data in CHUNKS)
        )
        httpd.terminate()

    def test_response(self):
        (httpd, client) = self.build_with_app(response_app)
        conn = client.connect()

        response = conn.request('GET', '/foo', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'content-length': len(DATA)})
        self.assertIsInstance(response.body, base.bodies.Body)
        self.assertEqual(response.body.read(), DATA)

        response = conn.request('GET', '/bar', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'content-length': 0})
        self.assertIsInstance(response.body, base.bodies.Body)
        self.assertEqual(response.body.read(), b'')

        response = conn.request('GET', '/baz', {}, None)
        self.assertEqual(response.status, 404)
        self.assertEqual(response.reason, 'Not Found')
        self.assertEqual(response.headers, {})
        self.assertIsNone(response.body)

        response = conn.request('GET', '/foo', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'content-length': len(DATA)})
        self.assertIsInstance(response.body, base.bodies.Body)
        self.assertEqual(response.body.read(), DATA)
        httpd.terminate()

    def test_always_accept_connections(self):
        marker = os.urandom(16)
        app = AppWithConnectionHandler(marker, True)
        (httpd, client) = self.build_with_app(app)
        for i in range(17):
            conn = client.connect()
            for j in range(69):
                response = conn.request('GET', '/', {}, None)
                self.assertEqual(response.status, 200)
                self.assertEqual(response.reason, 'OK')
                self.assertEqual(response.headers, {'content-length': 16})
                self.assertIsInstance(response.body, base.bodies.Body)
                self.assertEqual(response.body.read(), marker)
            conn.close()
        httpd.terminate()

    def test_always_reject_connections(self):
        marker = os.urandom(16)
        app = AppWithConnectionHandler(marker, False)
        (httpd, client) = self.build_with_app(app)
        for i in range(17):
            conn = client.connect()
            with self.assertRaises(ConnectionError):
                conn.request('GET', '/', {}, None)
            self.assertIs(conn.closed, True)
        httpd.terminate()

    def test_ok_status(self):
        (httpd, client) = self.build_with_app(standard_harness_app,
                                              max_requests=600)
        conn = client.connect()
        # At no point should the connection be closed by the server:
        for status in range(100, 400):
            reason = random_id()
            uri = '/status/{}/{}'.format(status, reason)
            response = conn.request('GET', uri, {}, None)
            self.assertEqual(response.status, status)
            self.assertEqual(response.reason, reason)
            self.assertEqual(response.headers, {})
            self.assertIsNone(response.body)
            # Again with a 2nd random reason string:
            reason = random_id()
            uri = '/status/{}/{}'.format(status, reason)
            response = conn.request('GET', uri, {}, None)
            self.assertEqual(response.status, status)
            self.assertEqual(response.reason, reason)
            self.assertEqual(response.headers, {})
            self.assertIsNone(response.body)
        conn.close()
        httpd.terminate()

    def test_error_status(self):
        (httpd, client) = self.build_with_app(standard_harness_app)
        for status in range(400, 600):
            reason = random_id()
            uri = '/status/{}/{}'.format(status, reason)
            conn = client.connect()
            response = conn.request('GET', uri, {}, None)
            self.assertEqual(response.status, status)
            self.assertEqual(response.reason, reason)
            self.assertEqual(response.headers, {})
            self.assertIsNone(response.body)
            if status in {404, 409, 412}:
                # Connection should not be closed for 404, 409, 412:
                response = conn.request('GET', uri, {}, None)
                self.assertEqual(response.status, status)
                self.assertEqual(response.reason, reason)
                self.assertEqual(response.headers, {})
                self.assertIsNone(response.body)
                conn.close()
            else:
                # But connection should be closed for all other status >= 400:
                with self.assertRaises(ConnectionError):
                    conn.request('GET', uri, {}, None)
                self.assertIs(conn.closed, True)
        httpd.terminate()

    def test_max_connections(self):
        uri = '/status/404/Nope'
        for value in (17, 27, 37):
            (httpd, client) = self.build_with_app(standard_harness_app,
                                                  max_connections=value)
            allconns = []
            for i in range(value):
                conn = client.connect()
                allconns.append(conn)
                response = conn.request('GET', uri, {}, None)
                self.assertEqual(response.status, 404)
                self.assertEqual(response.reason, 'Nope')
                self.assertEqual(response.headers, {})
                self.assertIsNone(response.body)
            with self.assertRaises((ConnectionError, ssl.SSLError)):
                conn = client.connect()
                conn.request('GET', uri, {}, None)
            for conn in allconns:
                conn.close()
            httpd.terminate()

    def test_max_requests(self):
        uri = '/status/404/Nope'
        for value in (17, 27, 37):
            (httpd, client) = self.build_with_app(standard_harness_app,
                                                  max_requests=value)
            conn = client.connect()
            for i in range(value):
                response = conn.request('GET', uri, {}, None)
                self.assertEqual(response.status, 404)
                self.assertEqual(response.reason, 'Nope')
                self.assertEqual(response.headers, {})
                self.assertIsNone(response.body)
            with self.assertRaises(ConnectionError):
                conn.request('GET', uri, {}, None)
            self.assertIs(conn.closed, True)
            httpd.terminate()


class TestLiveServer_AF_INET6(TestLiveServer):
    address = degu.IPv6_LOOPBACK


class TestLiveServer_AF_UNIX(TestLiveServer):
    def build_with_app(self, app, **options):
        tmp = TempDir()
        filename = tmp.join('my.socket')
        httpd = TempServer(filename, rgi.Validator(app), **options)
        httpd.tmp = tmp
        return (httpd, Client(httpd.address))


def ssl_app(session, request, bodies):
#    assert session['ssl_cipher'] == (
#        'ECDHE-RSA-AES128-GCM-SHA256', 'TLSv1/SSLv3', 128
#    )
#    assert session['ssl_compression'] is None
    assert request.method == 'GET'
    assert request.mount == []
    assert request.body is None
    return (200, 'OK', {}, None)


class TestLiveSSLServer(TestLiveServer):
    def build_with_app(self, app, **options):
        pki = TempPKI()
        httpd = TempSSLServer(
            pki.server_sslconfig, self.address, rgi.Validator(app), **options
        )
        httpd.pki = pki
        sslctx = build_client_sslctx(pki.client_sslconfig)
        return (httpd, SSLClient(sslctx, httpd.address))

    def test_ssl(self):
        pki = TempPKI(client_pki=True)
        server_config = pki.server_sslconfig
        client_config = pki.client_sslconfig
        httpd = TempSSLServer(server_config, self.address,  ssl_app)

        # Test from a non-SSL client:
        client = Client(httpd.address)
        conn = client.connect()
        with self.assertRaises(ConnectionResetError) as cm:
            conn.request('GET', '/', {}, None)
        self.assertEqual(str(cm.exception), '[Errno 104] Connection reset by peer')
        self.assertIs(conn.closed, True)

        # Test with no client cert:
        sslctx = build_client_sslctx({'ca_file': client_config['ca_file']})
        client = SSLClient(sslctx, httpd.address)
        with self.assertRaises(ssl.SSLError) as cm:
            client.connect()
        self.assertTrue(
            str(cm.exception).startswith('[SSL: SSLV3_ALERT_HANDSHAKE_FAILURE]')
        )
        self.assertIs(conn.closed, True)

        # Test with the wrong client cert (not signed by client CA):
        sslctx = build_client_sslctx({
            'ca_file': client_config['ca_file'],
            'cert_file': server_config['cert_file'],
            'key_file': server_config['key_file'],
        })
        client = SSLClient(sslctx, httpd.address)
        with self.assertRaises(ssl.SSLError) as cm:
            client.connect()
        self.assertTrue(
            str(cm.exception).startswith('[SSL: TLSV1_ALERT_UNKNOWN_CA]')
        )
        self.assertIs(conn.closed, True)

        # Test with a properly configured SSLClient:
        sslctx = build_client_sslctx(client_config)
        client = SSLClient(sslctx, httpd.address)
        conn = client.connect()
        response = conn.request('GET', '/', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertIsNone(response.body)

        # Test when check_hostname is True:
        conn.close()
        client.sslctx.check_hostname = True
        with self.assertRaises(ssl.CertificateError) as cm:
            client.connect()

        httpd.terminate()


class TestLiveSSLServer_AF_INET6(TestLiveSSLServer):
    address = degu.IPv6_LOOPBACK




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests the `degu.tables` module.
"""

from unittest import TestCase

from degu import tables


class TestConstants(TestCase):
    def test_NEVER_ALLOWED(self):
        self.assertIsInstance(tables.NEVER_ALLOWED, bytes)
        self.assertEqual(len(tables.NEVER_ALLOWED), 161)
        self.assertEqual(tables.NEVER_ALLOWED,
            tables.normalize(tables.NEVER_ALLOWED)
        )
        for b in tables.NEVER_ALLOWED:
            if b < 128:
                self.assertFalse(chr(b).isprintable())

    def test_NEVER_ALLOWED_SET(self):
        self.assertIsInstance(tables.NEVER_ALLOWED_SET, frozenset)
        self.assertEqual(tables.NEVER_ALLOWED_SET,
            frozenset(tables.NEVER_ALLOWED)
        )
        self.assertTrue(
            tables.NEVER_ALLOWED_SET.issuperset(range(32))
        )
        self.assertTrue(
            tables.NEVER_ALLOWED_SET.issuperset(range(127, 256))
        )

    def check_allowed(self, allowed):
        self.assertIsInstance(allowed, bytes)
        self.assertEqual(len(allowed), len(set(allowed)))
        self.assertEqual(allowed, bytes(sorted(set(allowed))))
        for i in range(256):
            if not chr(i).isprintable():
                self.assertNotIn(i, allowed)
        for i in allowed:
            self.assertEqual(i & 128, 0)

    def test_NAMES_DEF(self):
        self.check_allowed(tables.NAMES_DEF)
        self.assertEqual(min(tables.NAMES_DEF), ord('-'))
        self.assertEqual(max(tables.NAMES_DEF), ord('z'))
        self.assertEqual(len(tables.NAMES_DEF), 63)

    def test_BIT_FLAGS_DEF(self):
        self.assertIsInstance(tables.BIT_FLAGS_DEF, tuple)
        self.assertEqual(len(tables.BIT_FLAGS_DEF), 7)
        for item in tables.BIT_FLAGS_DEF:
            self.assertIsInstance(item, tuple)
            self.assertEqual(len(item), 2)
            (name, allowed) = item
            self.assertIsInstance(name, str)
            self.assertGreater(len(name), 1)
            self.assertTrue(name.isupper())
            self.assertTrue(name.isidentifier())
            self.check_allowed(allowed)

    def test_BIT_MASKS_DEF(self):
        self.assertIsInstance(tables.BIT_MASKS_DEF, tuple)
        self.assertGreaterEqual(len(tables.BIT_MASKS_DEF), 6)
        avail_flag_names = frozenset(dict((tables.BIT_FLAGS_DEF)))
        for item in tables.BIT_MASKS_DEF:
            self.assertIsInstance(item, tuple)
            self.assertEqual(len(item), 2)
            (name, flag_names) = item
            self.assertIsInstance(name, str)
            self.assertGreater(len(name), 1)
            self.assertTrue(name.isupper())
            self.assertTrue(name.isidentifier())
            self.assertIsInstance(flag_names, tuple)
            self.assertGreaterEqual(len(flag_names), 1)
            self.assertTrue(avail_flag_names.issuperset(flag_names))
            for n in flag_names:
                self.assertIsInstance(n, str)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests the `degu.sslhelpers` module.
"""

from unittest import TestCase
import os
from os import path

from .helpers import TempDir
from degu.sslhelpers import random_id
from degu import sslhelpers


class TestSSLFunctions(TestCase):
    def test_create_key(self):
        # 1024 bit:
        key_data = sslhelpers.create_key(1024)
        self.assertIn(len(key_data), [883, 887, 891])
        sslhelpers.get_pubkey(key_data)

        # 2048 bit:
        key_data = sslhelpers.create_key(2048)
        self.assertIn(len(key_data), [1671, 1675, 1679])
        sslhelpers.get_pubkey(key_data)

        # 3072 bit:
        key_data = sslhelpers.create_key(3072)
        self.assertIn(len(key_data), [2455, 2459])
        sslhelpers.get_pubkey(key_data)

        # 4096 bit:
        key_data = sslhelpers.create_key(4096)
        self.assertIn(len(key_data), [3239, 3243, 3247])
        sslhelpers.get_pubkey(key_data)

    def test_create_ca(self):
        key_data = sslhelpers.create_key(1024)
        pubkey_data = sslhelpers.get_pubkey(key_data)
        tmp = TempDir()
        key_file = tmp.write(key_data, 'foo.key')
        ca_data = sslhelpers.create_ca(key_file, '/CN=foo')
        self.assertEqual(sslhelpers.get_cert_pubkey(ca_data), pubkey_data)

    def test_create_csr(self):
        key = sslhelpers.create_key(1024)
        pubkey = sslhelpers.get_pubkey(key)
        tmp = TempDir()
        key_file = tmp.write(key, 'foo.key')
        csr = sslhelpers.create_csr(key_file, '/CN=foo')
        self.assertEqual(sslhelpers.get_csr_pubkey(csr), pubkey)

    def test_issue_cert(self):
        tmp = TempDir()

        foo_key_file = tmp.join('foo.key')
        foo_ca_file = tmp.join('foo.ca')
        foo_srl_file = tmp.join('foo.srl')
        foo_key = sslhelpers.create_key(1024)
        sslhelpers.safe_write(foo_key_file, foo_key)
        foo_ca = sslhelpers.create_ca(foo_key_file, '/CN=foo')
        sslhelpers.safe_write(foo_ca_file, foo_ca)

        bar_key_file = tmp.join('bar.key')
        bar_csr_file = tmp.join('bar.csr')
        bar_key = sslhelpers.create_key(1024)
        sslhelpers.safe_write(bar_key_file, bar_key)
        bar_csr = sslhelpers.create_csr(bar_key_file, '/CN=bar')
        sslhelpers.safe_write(bar_csr_file, bar_csr)

        bar_cert = sslhelpers.issue_cert(
            bar_csr_file, foo_ca_file, foo_key_file, foo_srl_file
        )
        self.assertEqual(
            sslhelpers.get_cert_pubkey(bar_cert),
            sslhelpers.get_pubkey(bar_key)
        )


class TestPKI(TestCase):
    def test_init(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        self.assertIs(pki.ssldir, tmp.dir)

    def test_path(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        cert_id = random_id(25)
        self.assertEqual(
            pki.path(cert_id, 'key'),
            tmp.join(cert_id + '.key')
        )
        self.assertEqual(
            pki.path(cert_id, 'cert'),
            tmp.join(cert_id + '.cert')
        )
        self.assertEqual(
            pki.path(cert_id, 'srl'),
            tmp.join(cert_id + '.srl')
        )

    def test_create_key(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        _id = pki.create_key(bits=1024)
        self.assertEqual(os.listdir(pki.ssldir), [ _id + '.key'])
        key_file = path.join(pki.ssldir, _id + '.key')
        key_data = open(key_file, 'rb').read()
        pubkey_data = sslhelpers.get_pubkey(key_data)
        self.assertEqual(_id, sslhelpers.hash_pubkey(pubkey_data))

    def test_create_ca(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        _id = pki.create_key(bits=1024)
        ca_file = tmp.join(_id + '.ca')
        self.assertFalse(path.exists(ca_file))
        self.assertEqual(pki.create_ca(_id), ca_file)
        self.assertTrue(path.isfile(ca_file))
        self.assertEqual(
            set(os.listdir(pki.ssldir)),
            set([_id + '.key', _id + '.ca'])
        )

    def test_create_csr(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        _id = pki.create_key(bits=1024)
        csr_file = tmp.join(_id + '.csr')
        self.assertFalse(path.exists(csr_file))
        self.assertEqual(pki.create_csr(_id), csr_file)
        self.assertTrue(path.isfile(csr_file))
        self.assertEqual(
            set(os.listdir(pki.ssldir)),
            set([_id + '.key', _id + '.csr'])
        )

    def test_issue_cert(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)

        # Create the CA
        ca_id = pki.create_key(bits=1024)
        pki.create_ca(ca_id)

        # Create the CSR
        cert_id = pki.create_key(bits=1024)
        pki.create_csr(cert_id)
        os.remove(tmp.join(cert_id + '.key'))

        # Now test
        cert_file = tmp.join(cert_id + '.cert')
        self.assertFalse(path.exists(cert_file))
        self.assertEqual(pki.issue_cert(cert_id, ca_id), cert_file)
        self.assertGreater(path.getsize(cert_file), 0)
        self.assertEqual(
            set(os.listdir(pki.ssldir)),
            set([
                ca_id + '.key',
                ca_id + '.ca',
                ca_id + '.srl',
                cert_id + '.csr',
                cert_id + '.cert',
            ])
        )

    def test_get_server_sslconfig(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        server_id = random_id()
        ca_id = random_id()
        self.assertEqual(
            pki.get_server_sslconfig(server_id, ca_id),
            {
                'cert_file': pki.path(server_id, 'cert'),
                'key_file': pki.path(server_id, 'key'),
                'ca_file': pki.path(ca_id, 'ca'),
            }
        )

    def test_get_anonymous_server_sslconfig(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        server_id = random_id()
        self.assertEqual(
            pki.get_anonymous_server_sslconfig(server_id),
            {
                'cert_file': pki.path(server_id, 'cert'),
                'key_file': pki.path(server_id, 'key'),
                'allow_unauthenticated_clients': True,
            }
        )

    def test_get_client_sslconfig(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        ca_id = random_id()
        client_id = random_id()
        self.assertEqual(
            pki.get_client_sslconfig(ca_id, client_id),
            {
                'ca_file': pki.path(ca_id, 'ca'),
                'check_hostname': False,
                'cert_file': pki.path(client_id, 'cert'),
                'key_file': pki.path(client_id, 'key'),
            }
        )

    def test_get_anonymous_client_sslconfig(self):
        tmp = TempDir()
        pki = sslhelpers.PKI(tmp.dir)
        ca_id = random_id()
        self.assertEqual(
            pki.get_anonymous_client_sslconfig(ca_id),
            {
                'ca_file': pki.path(ca_id, 'ca'),
                'check_hostname': False,
            }
        )




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu._basepy` module.
"""

from unittest import TestCase
from random import SystemRandom
import os

from ..sslhelpers import random_id
from .._basepy import TYPE_ERROR, TYPE_ERROR2
from .. import _basepy


random = SystemRandom()


class TestFunctions(TestCase):
    def test_check_type(self):
        _check_type = _basepy._check_type

        # Static value tests:
        self.assertIsNone(_check_type('foo', 17, int))
        with self.assertRaises(TypeError) as cm:
            _check_type('foo', 17.0, int)
        self.assertEqual(str(cm.exception),
            "foo: need a <class 'int'>; got a <class 'float'>: 17.0"
        )

        self.assertIsNone(_check_type('bar', 'hello', str))
        with self.assertRaises(TypeError) as cm:
            _check_type('bar', b'hello', str)
        self.assertEqual(str(cm.exception),
            "bar: need a <class 'str'>; got a <class 'bytes'>: b'hello'"
        )

        self.assertIsNone(_check_type('baz', b'hello', bytes))
        with self.assertRaises(TypeError) as cm:
            _check_type('baz', 'hello', bytes)
        self.assertEqual(str(cm.exception),
            "baz: need a <class 'bytes'>; got a <class 'str'>: 'hello'"
        )

        self.assertIsNone(_check_type('car', [], list))
        with self.assertRaises(TypeError) as cm:
            _check_type('car', {}, list)
        self.assertEqual(str(cm.exception),
            "car: need a <class 'list'>; got a <class 'dict'>: {}"
        )

        self.assertIsNone(_check_type('cdr', {}, dict))
        with self.assertRaises(TypeError) as cm:
            _check_type('cdr', [], dict)
        self.assertEqual(str(cm.exception),
            "cdr: need a <class 'dict'>; got a <class 'list'>: []"
        )

        # Radom value tests:
        values = (
            0,
            random.randrange(-999, 0),
            random.randrange(1, 1001),

            '',
            random_id(15),
            random_id(30),

            b'',
            os.urandom(15),
            os.urandom(30),

            [],
            [random_id()],

            {},
            {random_id(): random_id()},
        )
        for obj in values:
            name = random_id()
            _type = type(obj)
            self.assertIsNone(_check_type(name, obj, _type))

            class subtype(_type):
                pass

            bad = subtype(obj)
            with self.assertRaises(TypeError) as cm:
                _check_type(name, bad, _type)
            self.assertEqual(str(cm.exception),
                TYPE_ERROR.format(name, _type, subtype, bad)
            )

    def test_check_type2(self):
        _check_type2 = _basepy._check_type2

        # Static value tests:
        self.assertIsNone(_check_type2('foo', 17, int))
        with self.assertRaises(TypeError) as cm:
            _check_type2('foo', 17.0, int)
        self.assertEqual(str(cm.exception),
            "foo: need a <class 'int'>; got a <class 'float'>"
        )

        self.assertIsNone(_check_type2('bar', 'hello', str))
        with self.assertRaises(TypeError) as cm:
            _check_type2('bar', b'hello', str)
        self.assertEqual(str(cm.exception),
            "bar: need a <class 'str'>; got a <class 'bytes'>"
        )

        self.assertIsNone(_check_type2('baz', b'hello', bytes))
        with self.assertRaises(TypeError) as cm:
            _check_type2('baz', 'hello', bytes)
        self.assertEqual(str(cm.exception),
            "baz: need a <class 'bytes'>; got a <class 'str'>"
        )

        self.assertIsNone(_check_type2('car', [], list))
        with self.assertRaises(TypeError) as cm:
            _check_type2('car', {}, list)
        self.assertEqual(str(cm.exception),
            "car: need a <class 'list'>; got a <class 'dict'>"
        )

        self.assertIsNone(_check_type2('cdr', {}, dict))
        with self.assertRaises(TypeError) as cm:
            _check_type2('cdr', [], dict)
        self.assertEqual(str(cm.exception),
            "cdr: need a <class 'dict'>; got a <class 'list'>"
        )

        # Radom value tests:
        values = (
            0,
            random.randrange(-999, 0),
            random.randrange(1, 1001),

            '',
            random_id(15),
            random_id(30),

            b'',
            os.urandom(15),
            os.urandom(30),

            [],
            [random_id()],

            {},
            {random_id(): random_id()},
        )
        for obj in values:
            name = random_id()
            _type = type(obj)
            self.assertIsNone(_check_type2(name, obj, _type))

            class subtype(_type):
                pass

            bad = subtype(obj)
            with self.assertRaises(TypeError) as cm:
                _check_type2(name, bad, _type)
            self.assertEqual(str(cm.exception),
                TYPE_ERROR2.format(name, _type, subtype)
            )

    def test_check_tuple(self):
        _check_tuple = _basepy._check_tuple
        values = tuple(
            tuple(random_id() for i in range(count))
            for count in range(10)
        )
        for obj in values:
            name = random_id()
            size = len(obj)
            self.assertIs(_check_tuple(name, obj, size), obj)
            bad = list(obj)
            with self.assertRaises(TypeError) as cm:
                _check_tuple(name, bad, size)
            self.assertEqual(str(cm.exception),
                TYPE_ERROR2.format(name, tuple, list)
            )
            for offset in (-2, -1, 1, 2):
                s = size + offset
                with self.assertRaises(ValueError) as cm:
                    _check_tuple(name, obj, s)
                self.assertEqual(str(cm.exception),
                    '{}: need a {}-tuple; got a {}-tuple'.format(name, s, size)
                )

    def test_check_list(self):
        _check_list = _basepy._check_list
        values = tuple(
            [random_id() for i in range(count)]
            for count in range(10)
        )
        for obj in values:
            name = random_id()
            self.assertIs(_check_list(name, obj), obj)
            bad = tuple(obj)
            with self.assertRaises(TypeError) as cm:
                _check_list(name, bad)
            self.assertEqual(str(cm.exception),
                TYPE_ERROR2.format(name, list, tuple)
            )




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Run the `degu` unit tests.
"""

import sys
import os
from os import path
import stat
from unittest import TestLoader, TextTestRunner
from doctest import DocTestSuite

import degu


packagedir = path.dirname(path.abspath(degu.__file__))


def pynames_iter(pkdir=packagedir, pkname=None):
    """
    Recursively yield dotted names for *.py files in directory *pydir*.
    """
    if not path.isfile(path.join(pkdir, '__init__.py')):
        return
    if pkname is None:
        pkname = path.basename(pkdir)
    yield pkname
    dirs = []
    for name in sorted(os.listdir(pkdir)):
        if name in ('__init__.py', '__pycache__'):
            continue
        if name.startswith('.') or name.endswith('~'):
            continue
        fullname = path.join(pkdir, name)
        st = os.lstat(fullname)
        if stat.S_ISREG(st.st_mode) and name.endswith('.py'):
            parts = name.split('.')
            if len(parts) == 2:
                yield '.'.join([pkname, parts[0]])
        elif stat.S_ISDIR(st.st_mode):
            dirs.append((fullname, name))
    for (fullname, name) in dirs:
        subpkg = '.'.join([pkname, name])
        for n in pynames_iter(fullname, subpkg):
            yield n


def run_tests():
    pynames = tuple(pynames_iter())

    # Add unit-tests:
    loader = TestLoader()
    suite = loader.loadTestsFromNames(pynames)

    # Add doc-tests:
    for name in pynames:
        suite.addTest(DocTestSuite(name))

    # Run the tests:
    runner = TextTestRunner(verbosity=2)
    result = runner.run(suite)
    print(
        'degu: {!r}'.format(path.abspath(degu.__file__)),
        file=sys.stderr
    )
    print('-' * 70, file=sys.stderr)
    return result.wasSuccessful()


if __name__ == '__main__':
    if not run_tests():
        raise SystemExit(2)




#!/usr/bin/python3

import time
import logging
import json
import argparse
import statistics
import platform
import sys

import degu
from degu.sslhelpers import random_id
from degu.misc import TempServer
from degu.tests.helpers import TempDir
from degu.client import Client


REQUESTS = 10000
RUNS = 20

parser = argparse.ArgumentParser()
parser.add_argument('--json', action='store_true', default=False,
    help='Output result in machine-readable JSON'
)
parser.add_argument('--send-host', action='store_true', default=False,
    help='Send an HTTP Host header for AF_INET6'
)
parser.add_argument('--unix', action='store_true', default=False,
    help='Use AF_UNIX instead of AF_INET6'
)
parser.add_argument('--requests', type=int, metavar='N', default=REQUESTS,
    help='Number of requests per run; default={}'.format(REQUESTS)
)
parser.add_argument('--runs', type=int, metavar='N', default=RUNS,
    help='Number of runs; default={}'.format(RUNS)
)
args = parser.parse_args()


logging.basicConfig(
    level=logging.DEBUG,
    format='\t'.join([
        '%(levelname)s',
        '%(threadName)s',
        '%(message)s',
    ]),
)

ping = random_id(60)
request_body = json.dumps({'ping': ping}).encode()
pong = random_id(60)
response_body = json.dumps({'pong': pong}).encode()


def ping_pong_app(session, request, bodies):
    request.body.read()
    return (200, 'OK', {}, response_body)


if args.unix:
    tmp = TempDir()
    address = tmp.join('my.socket')
else:
    address = degu.IPv6_LOOPBACK
server = TempServer(address, ping_pong_app, max_requests=args.requests)
if args.send_host:
    client = Client(server.address)
else:
    client = Client(server.address, host=None)


deltas = []
for i in range(args.runs):
    conn = client.connect()
    start = time.monotonic()
    for i in range(args.requests):
        conn.post('/', {}, request_body).body.read()
    deltas.append(time.monotonic() - start)
    conn.close()
del conn
del server

rates = tuple(args.requests / d for d in deltas)
_max = max(rates)
_mean = statistics.mean(rates)
_min = min(rates)
_stdev = statistics.stdev(rates)
pyinfo = '{}; {}; {} ({} {})'.format(
    platform.python_version(),
    platform.machine(),
    platform.system(),
    platform.dist()[0],
    platform.dist()[1],
)
family = ('AF_UNIX' if args.unix else 'AF_INET6')
if args.json:
    doc = {
        'Degu': degu.__version__,
        'Python': pyinfo,
        'family': family,
        'requests': args.requests,
        'runs': args.runs,
        'fastest': _max,
        'average': _mean,
        'slowest': _min,
        'stdev': _stdev,
    }
    print(json.dumps(doc, sort_keys=True, indent=4))
    sys.exit(0)

fastest_run = str(rates.index(_max) + 1)
slowest_run = str(rates.index(_min) + 1)
width1 = max(len(s) for s in [fastest_run, slowest_run])

fastest = '{:,.0f}'.format(_max)
mean = '{:,.0f}'.format(_mean)
slowest = '{:,.0f}'.format(_min)
stdev = '{:,.0f}'.format(_stdev)
width2 = max(len(s) for s in [fastest, mean, slowest, stdev])

print('')
print('Degu: {}'.format(degu.__version__))
print('Python: {}'.format(pyinfo))
print('Test: {}; {:,} requests per run; {} runs'.format(
    family, args.requests, args.runs)
)
print('Request body: {:d} bytes'.format(len(request_body)))
print('Response body: {:d} bytes'.format(len(response_body)))
print('-' * 72)
print('Run {} of {} was fastest'.format(fastest_run.rjust(width1), args.runs))
print('Run {} of {} was slowest'.format(slowest_run.rjust(width1), args.runs))
print('Requests per second:')
print('    fastest: {}'.format(fastest.rjust(width2)))
print('    average: {}'.format(mean.rjust(width2)))
print('    slowest: {}'.format(slowest.rjust(width2)))
print('      stdev: {}'.format(stdev.rjust(width2)))
print('-' * 72)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Common HTTP parser and IO abstractions used by server and client.
"""

try:
    from ._base import (
        EmptyPreambleError,
        Bodies,   BodiesType,
        Response, ResponseType,
        Range,
        ContentRange,
        Request,
        bodies,
        Connection,
        Session,
        _handle_requests,
        parse_headers,
    )
except ImportError:
    from ._basepy import (
        EmptyPreambleError,
        Bodies,   BodiesType,
        Response, ResponseType,
        Range,
        ContentRange,
        Request,
        bodies,
        Connection,
        Session,
        _handle_requests,
        parse_headers,
    )


__all__ = (
    'EmptyPreambleError',
    'Bodies', 'BodiesType',
    'Response', 'ResponseType',
    'Range',
    'ContentRange',
    'Request',
    'bodies',
    'Connection',
    'Session',
    '_handle_requests',
    'parse_headers',
)

MAX_READ_SIZE = 16777216  # 16 MiB
MAX_CHUNK_SIZE = 16777216  # 16 MiB
IO_SIZE = 1048576  # 1 MiB

# Provide very clear TypeError messages:
_TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'


# FIXME: Add optional max_size=None keyword argument
def read_chunk(rfile):
    """
    Read a chunk from a chunk-encoded request or response body.

    See "Chunked Transfer Coding":

        http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.6.1
    """
    line = rfile.readline(4096)
    if line[-2:] != b'\r\n':
        raise ValueError('bad chunk size termination: {!r}'.format(line[-2:]))
    parts = line[:-2].split(b';')
    if len(parts) > 2:
        raise ValueError('bad chunk size line: {!r}'.format(line))
    size = int(parts[0], 16)
    if not (0 <= size <= MAX_CHUNK_SIZE):
        raise ValueError(
            'need 0 <= chunk_size <= {}; got {}'.format(MAX_CHUNK_SIZE, size)
        )
    if len(parts) == 2:
        text = None
        try:
            text = parts[1].decode('ascii')  # Disallow the high-bit
        except ValueError:
            pass
        if text is None or not text.isprintable():
            raise ValueError(
                'bad bytes in chunk extension: {!r}'.format(parts[1])
            )
        (key, value) = text.split('=')
        extension = (key, value)
    else:
        extension = None
    data = rfile.read(size)
    if len(data) != size:
        raise ValueError('underflow: {} < {}'.format(len(data), size))
    crlf = rfile.read(2)
    if crlf != b'\r\n':
        raise ValueError('bad chunk data termination: {!r}'.format(crlf))
    return (extension, data)


def _encode_chunk(chunk, check_size=True):
    """
    Internal API for unit testing.
    """
    assert isinstance(chunk, tuple)
    (extension, data) = chunk
    assert extension is None or isinstance(extension, tuple)
    assert isinstance(data, bytes)
    if check_size and len(data) > MAX_CHUNK_SIZE:
        raise ValueError(
            'need len(data) <= {}; got {}'.format(MAX_CHUNK_SIZE, len(data))
        )
    if extension is None:
        size_line = '{:x}\r\n'.format(len(data))
    else:
        (key, value) = extension
        assert isinstance(key, str)
        assert isinstance(value, str)
        size_line = '{:x};{}={}\r\n'.format(len(data), key, value)
    return b''.join([size_line.encode(), data, b'\r\n'])


def write_chunk(wfile, chunk, max_size=None):
    """
    Write *chunk* to *wfile* using chunked transfer-encoding.

    Warning: the optional *max_size* keyword argument isn't yet part of the
    stable API, might go away or change in behavior.

    See "Chunked Transfer Coding":

        http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.6.1
    """
    assert isinstance(chunk, tuple)
    (extension, data) = chunk
    assert extension is None or isinstance(extension, tuple)
    assert isinstance(data, bytes)
    if max_size is None:
        max_size = MAX_CHUNK_SIZE
    if len(data) > max_size:
        raise ValueError(
            'need len(data) <= {}; got {}'.format(max_size, len(data))
        )
    if extension is None:
        size_line = '{:x}\r\n'.format(len(data))
    else:
        (key, value) = extension
        assert isinstance(key, str)
        assert isinstance(value, str)
        size_line = '{:x};{}={}\r\n'.format(len(data), key, value)
    total = wfile.write(b''.join([size_line.encode(), data, b'\r\n']))
    # Flush buffer as it could be some time before the next chunk is available:
    wfile.flush()
    return total


def format_headers(headers, sort=True):
    lines = ['{}: {}'.format(*kv) for kv in headers.items()]
    if sort:
        lines.sort()
    return '\r\n'.join(lines).encode()

